################################################################################
#                                                                              #
#  R functions for the monotone single index hazard model (Qin et al., 2020+)  #
#                                                                              #
################################################################################
# SIM() is the main function to calculate the solution of beta, r,
# and \Lambda. See Example.R for an example on how to use the function.
#
# Arguments
#   ldata - The data frame of observed data. The first p columns are p covariates,
# and the last two columns are the observed survival time and binary variable
# indicating whether the observation is censored (0) or not (1). The names of the
# last two columns must be "Y" and "E"
#   K -  The number of additional initial values to try. If K=0, only the initial
#   value calculated from the Cox model is used
#   zz - The vector at which the estimated r function will be evaluated
#   tt - The vector at which the estimated Lambda function will be evaluated
#
# Values
# A list with the following components:
# beta - The estimated beta
# betacox - The normalized coefficients from the Cox model, also used to calculate
# the first initial value of the proposed algorithm
# Ltt - The estimated Lambda function evaluated at the vector tt
# Rzz - The estimated r function evaluated at the vector zz
################################################################################
# Other functions called by SIM()
# L3_new evaluates the profile likelihood for a given theta
# trans/trans2 are used to transform between theta (p-1 dimensional vector) and
# beta (p dimentional vector with unit norm)
# PhiLam calculate r and Lambda at a user specified grid after obtaining MLE
################################################################################


SIM <- function(ldata, K, zz, tt)
{
  P <- dim(ldata)[2]-2
  ldata$Y <- round(ldata$Y, 10)
  ldata <- orderBy(~Y+E, data=ldata)
  N <- dim(ldata)[1]
  ldata$IND <- 1:N

  yy1 <- unique(ldata$Y)

  ycut <- as.integer(cut(c(ldata$Y),
                         c(-10^10,yy1, 10^10),
                         dig.lab = 12))

  Eyy1 <- ctapply(ldata$E, ycut, sum)
  N1 <- length(yy1)



  fitcox <- coxph(Surv(Y,E)~., data = ldata[,1:(P+2)])
  estcox <- fitcox$coefficients/sqrt(sum(fitcox$coefficients^2))

  opt0 <- optim(trans(-estcox), L3_new, ldata = ldata, N1 = N1, N = N, P = P,
                ycut = ycut, Eyy1 = Eyy1, yy1 = yy1,
                control = list(warn.1d.NelderMead = FALSE))

  if(K == 0)
  {
    theta <- (opt0$par)
  }else{

    init0 <- opt0$par
    va0 <- opt0$value

    for(j in 1:K)
    {
      init <- init0+runif(P-1,-1,1)*.5
      so <- optim(init, L3_new, ldata = ldata, N1 = N1, N = N, P = P,
                  ycut = ycut, Eyy1 = Eyy1, yy1 = yy1,
                  control = list(warn.1d.NelderMead = FALSE))
      va <- so$value
      if(va<va0)
      {
        init0 <- so$par
        va0 <- va
      }
    }
    theta <- (init0)
  }

  pm <- PhiLam(zz, tt, theta, ldata, N1, N, P, ycut, Eyy1, yy1)
  list(beta = trans2(theta),
       betacox = estcox,
       Ltt = pm$Ltt,
       rzz = pm$rzz)
}

L3_new <- function(theta,ldata,N1,N,P,ycut,Eyy1,yy1)
{
  beta <- cumprod(c(1,sin(theta)))*c(cos(theta),1)
  ldata$Z <- round((as.matrix(ldata)[,1:P])%*%beta,10)
  ldata2 <- orderBy(~Z+E, data=ldata)
  zz1 <- (unique(ldata2$Z[ldata2$E==1]))
  zcut <- as.integer(cut(c(ldata2$Z),
                         c(-10^10,zz1, 10^10),
                         dig.lab = 12))

  foo2 <- ctapply(ldata2$E, zcut, sum)
  odlz <- ldata2$IND
  tz <- as.vector(table(zcut))

  iterationPhiLam <- function(par)
  {
    phiold <- par[1:N]
    Lamold <- par[(N+1):(2*N)]
    Lamold2 <- Lamold[odlz]
    foo1 <- ctapply(Lamold2, zcut, sum)

    eta <- foo1/foo2

    if( any(foo2==0) )
    {
      pout <- pava(y=eta[foo2!=0], w=foo2[foo2!=0])
      pout <- c(pout, max(pout))
    } else{
      pout <- pava(y=eta, w=foo2)
    }

    pout[pout<0] <- 0
    rnew <- rep(pout, tz)
    phinew <- -log( rnew )

    ldata$rnew <- 1
    ldata$rnew[ldata2$IND] <- rnew

    ryy1 <- ctapply(1/ldata$rnew, ycut, sum)
    rsumyy1 <- rep(0, N1)
    rsumyy1[N1:1] <- cumsum(ryy1[N1:1])
    Lamyy1 <- cumsum(Eyy1/rsumyy1)
    Lamtau <- max(Lamyy1)
    Lamnew <- Lamyy1[match(ldata$Y,yy1)]/Lamtau
    phinew <- phinew+log(Lamtau)


    c(phinew, Lamnew)
  }
  parnew2 <- squarem(par = c(rep(1,N), ldata$Y),
                     fixptfn = iterationPhiLam,
                     control = list())$par


  phinew <- parnew2[1:N]
  ldata$phinew <- 1
  ldata$phinew[ldata2$IND] <- phinew

  risk <- rep(0,N)
  risk[N:1] <- cumsum(exp(ldata$phinew[N:1]))

  lam <- 1/risk
  - sum(log(lam[ldata$E==1])) - sum(phinew[ldata2$E==1])
}

PhiLam <- function(zz,tt,
                   theta,ldata,N1,N,P,ycut,Eyy1,yy1)
{
  beta <- cumprod(c(1,sin(theta)))*c(cos(theta),1)
  ldata$Z <- round((as.matrix(ldata)[,1:P])%*%beta,10)
  ldata2 <- orderBy(~Z+E, data=ldata)
  zz1 <- (unique(ldata2$Z[ldata2$E==1]))
  zcut <- as.integer(cut(c(ldata2$Z),
                         c(-10^10,zz1, 10^10),
                         dig.lab = 12))

  foo2 <- ctapply(ldata2$E, zcut, sum)
  odlz <- ldata2$IND
  tz <- as.vector(table(zcut))

  iterationPhiLam <- function(par)
  {
    phiold <- par[1:N]
    Lamold <- par[(N+1):(2*N)]
    Lamold2 <- Lamold[odlz]
    foo1 <- ctapply(Lamold2, zcut, sum)

    eta <- foo1/foo2

    if( any(foo2==0) )
    {
      pout <- pava(y=eta[foo2!=0], w=foo2[foo2!=0])
      pout <- c(pout, max(pout))
    } else{
      pout <- pava(y=eta, w=foo2)
    }

    pout[pout<0] <- 0
    rnew <- rep(pout, tz)
    phinew <- -log( rnew )

    ldata$rnew <- 1
    ldata$rnew[ldata2$IND] <- rnew

    ryy1 <- ctapply(1/ldata$rnew, ycut, sum)
    rsumyy1 <- rep(0, N1)
    rsumyy1[N1:1] <- cumsum(ryy1[N1:1])
    Lamyy1 <- cumsum(Eyy1/rsumyy1)
    Lamtau <- max(Lamyy1)
    Lamnew <- Lamyy1[match(ldata$Y,yy1)]/Lamtau
    phinew <- phinew+log(Lamtau)


    c(phinew, Lamnew)
  }
  parnew2 <- squarem(par = c(rep(1,N), ldata$Y),
                     fixptfn = iterationPhiLam,
                     control = list())$par


  phinew <- parnew2[1:N]
  Lamnew <- parnew2[(N+1):(2*N)]

  L <- stepfun(x = c(ldata$Y),
               y = c(0,Lamnew))
  r <- stepfun(x = c(ldata2$Z),
               y = c(exp(-phinew),exp(-phinew[N])))

  list(Ltt = L(tt),
       rzz = r(zz))
}

trans <- function(beta)
{
  j <- 1
  stheta <- 1
  lb <- length(beta)
  theta <- rep(0, lb-1)
  while(j < lb)
  {
    theta[j] <- acos(beta[j]/stheta[j])
    stheta[j+1] <- stheta[j]*sin(theta[j])
    j <- j + 1
  }
  if(sin(theta[j-1])*(beta[j]/stheta[j])<0){
    theta[j-1] <- -abs(theta[j-1])
  }
  theta
}

trans2 <- function(theta)
{
  beta <- cumprod(c(1,sin(theta)))*c(cos(theta),1)
  beta
}
