################################################################################
# SIM() is the main function to calculate the solution of beta, r,
# and \Lambda. See Example.R for an example on how to use the function.
#
# Arguments
#   ldata - The data frame of observed data. The first p columns are p covariates,
# and the last two columns are the observed survival time and binary variable
# indicating whether the observation is censored (0) or not (1). The names of the
# last two columns must be "Y" and "E"
#   K -  The number of additional initial values to try. If K=0, only the initial
#   value calculated from the Cox model is used
#   zz - The vector at which the estimated r function will be evaluated
#   tt - The vector at which the estimated Lambda function will be evaluated
#
# Values
# A list with the following components:
# beta - The estimated beta
# betacox - The normalized coefficients from the Cox model, also used to calculate
# the first initial value of the proposed algorithm
# Ltt - The estimated Lambda function evaluated at the vector tt
# Rzz - The estimated r function evaluated at the vector zz
################################################################################
# Other functions called by SIM()
# L3_new evaluates the profile likelihood for a given theta
# trans/trans2 are used to transform between theta (p-1 dimensional vector) and
# beta (p dimentional vector with unit norm)
# PhiLam calculate r and Lambda at a user specified grid after obtaining MLE
################################################################################

L3_new <- function(theta,ldata,N1,N,P,ycut,Eyy1,yy1)
{
  beta <- cumprod(c(1,sin(theta)))*c(cos(theta),1)
  ldata$Z <- round((as.matrix(ldata)[,1:P])%*%beta,10)
  ldata2 <- orderBy(~Z+E, data=ldata)
  zz1 <- (unique(ldata2$Z[ldata2$E==1]))
  zcut <- as.integer(cut(c(ldata2$Z),
                         c(-10^10,zz1, 10^10),
                         dig.lab = 12))

  foo2 <- ctapply(ldata2$E, zcut, sum)
  odlz <- ldata2$IND
  tz <- as.vector(table(zcut))

  iterationPhiLam <- function(par)
  {
    phiold <- par[1:N]
    Lamold <- par[(N+1):(2*N)]
    Lamold2 <- Lamold[odlz]
    foo1 <- ctapply(Lamold2, zcut, sum)

    eta <- foo1/foo2

    if( any(foo2==0) )
    {
      pout <- pava(y=eta[foo2!=0], w=foo2[foo2!=0])
      pout <- c(pout, max(pout))
    } else{
      pout <- pava(y=eta, w=foo2)
    }

    pout[pout<0] <- 0
    rnew <- rep(pout, tz)
    phinew <- -log( rnew )

    ldata$rnew <- 1
    ldata$rnew[ldata2$IND] <- rnew

    ryy1 <- ctapply(1/ldata$rnew, ycut, sum)
    rsumyy1 <- rep(0, N1)
    rsumyy1[N1:1] <- cumsum(ryy1[N1:1])
    Lamyy1 <- cumsum(Eyy1/rsumyy1)
    Lamtau <- max(Lamyy1)
    Lamnew <- Lamyy1[match(ldata$Y,yy1)]/Lamtau
    phinew <- phinew+log(Lamtau)


    c(phinew, Lamnew)
  }
  parnew2 <- squarem(par = c(rep(1,N), ldata$Y),
                     fixptfn = iterationPhiLam,
                     control = list())$par


  phinew <- parnew2[1:N]
  ldata$phinew <- 1
  ldata$phinew[ldata2$IND] <- phinew

  risk <- rep(0,N)
  risk[N:1] <- cumsum(exp(ldata$phinew[N:1]))

  lam <- 1/risk
  - sum(log(lam[ldata$E==1])) - sum(phinew[ldata2$E==1])
}

PhiLam <- function(zz,tt,
                   theta,ldata,N1,N,P,ycut,Eyy1,yy1)
{
  beta <- cumprod(c(1,sin(theta)))*c(cos(theta),1)
  ldata$Z <- round((as.matrix(ldata)[,1:P])%*%beta,10)
  ldata2 <- orderBy(~Z+E, data=ldata)
  zz1 <- (unique(ldata2$Z[ldata2$E==1]))
  zcut <- as.integer(cut(c(ldata2$Z),
                         c(-10^10,zz1, 10^10),
                         dig.lab = 12))

  foo2 <- ctapply(ldata2$E, zcut, sum)
  odlz <- ldata2$IND
  tz <- as.vector(table(zcut))

  iterationPhiLam <- function(par)
  {
    phiold <- par[1:N]
    Lamold <- par[(N+1):(2*N)]
    Lamold2 <- Lamold[odlz]
    foo1 <- ctapply(Lamold2, zcut, sum)

    eta <- foo1/foo2

    if( any(foo2==0) )
    {
      pout <- pava(y=eta[foo2!=0], w=foo2[foo2!=0])
      pout <- c(pout, max(pout))
    } else{
      pout <- pava(y=eta, w=foo2)
    }

    pout[pout<0] <- 0
    rnew <- rep(pout, tz)
    phinew <- -log( rnew )

    ldata$rnew <- 1
    ldata$rnew[ldata2$IND] <- rnew

    ryy1 <- ctapply(1/ldata$rnew, ycut, sum)
    rsumyy1 <- rep(0, N1)
    rsumyy1[N1:1] <- cumsum(ryy1[N1:1])
    Lamyy1 <- cumsum(Eyy1/rsumyy1)
    Lamtau <- max(Lamyy1)
    Lamnew <- Lamyy1[match(ldata$Y,yy1)]/Lamtau
    phinew <- phinew+log(Lamtau)


    c(phinew, Lamnew)
  }
  parnew2 <- squarem(par = c(rep(1,N), ldata$Y),
                     fixptfn = iterationPhiLam,
                     control = list())$par


  phinew <- parnew2[1:N]
  Lamnew <- parnew2[(N+1):(2*N)]

  L <- stepfun(x = c(ldata$Y),
               y = c(0,Lamnew))
  r <- stepfun(x = c(ldata2$Z),
               y = c(exp(-phinew),exp(-phinew[N])))

  list(Ltt = L(tt),
       rzz = r(zz))
}

SIM <- function(ldata, K, zz, tt)
{
  P <- dim(ldata)[2]-2
  ldata$Y <- round(ldata$Y, 10)
  ldata <- orderBy(~Y+E, data=ldata)
  N <- dim(ldata)[1]
  ldata$IND <- 1:N

  yy1 <- unique(ldata$Y)

  ycut <- as.integer(cut(c(ldata$Y),
                         c(-10^10,yy1, 10^10),
                         dig.lab = 12))

  Eyy1 <- ctapply(ldata$E, ycut, sum)
  N1 <- length(yy1)



  fitcox <- coxph(Surv(Y,E)~., data = ldata[,1:(P+2)])
  estcox <- fitcox$coefficients/sqrt(sum(fitcox$coefficients^2))

  opt0 <- optim(trans(-estcox), L3_new, ldata = ldata, N1 = N1, N = N, P = P,
                ycut = ycut, Eyy1 = Eyy1, yy1 = yy1)

  if(K == 0)
  {
    theta <- (opt0$par)
  }else{

    init0 <- opt0$par
    va0 <- opt0$value

    for(j in 1:K)
    {
      init <- init0+runif(P-1,-1,1)*.5
      so <- optim(init, L3_new, ldata = ldata, N1 = N1, N = N, P = P,
                  ycut = ycut, Eyy1 = Eyy1, yy1 = yy1)
      va <- so$value
      if(va<va0)
      {
        init0 <- so$par
        va0 <- va
      }
    }
    theta <- (init0)
  }

  pm <- PhiLam(zz, tt, theta, ldata, N1, N, P, ycut, Eyy1, yy1)
  list(beta = trans2(theta),
       betacox = estcox,
       Ltt = pm$Ltt,
       rzz = pm$rzz)
}

trans <- function(beta)
{
  j <- 1
  stheta <- 1
  lb <- length(beta)
  theta <- rep(0, lb-1)
  while(j < lb)
  {
    theta[j] <- acos(beta[j]/stheta[j])
    stheta[j+1] <- stheta[j]*sin(theta[j])
    j <- j + 1
  }
  if(sin(theta[j-1])*(beta[j]/stheta[j])<0){
    theta[j-1] <- -abs(theta[j-1])
  }
  theta
}

trans2 <- function(theta)
{
  beta <- cumprod(c(1,sin(theta)))*c(cos(theta),1)
  beta
}

################################################################################
# Functions needed in simulations
# genData: generate simulated datasets
# calPara: calculate the parameters of censoring distribution to achieve the desired censoring rate
# do.sim: generate data and estimate the parameters using SIM()
################################################################################

genData <- function(cc, tau, sce = 1, cr0 = 0.75, N = 800, a = 1/3, res = 1)
{
  if(sce == 1)
  {
    P <- 2
    beta0 <- c(-1,1)
    X <- matrix(NA, ncol = P, nrow = N)
    X[,1] <- rexp(N, 1)
    X[,2] <- rexp(N, 1)
  }else if(sce == 2)
  {
    P <- 2
    beta0 <- c(-1,1)
    X <- matrix(NA, ncol = P, nrow = N)
    X[,1] <- rnorm(N)
    X[,2] <- rnorm(N)
  }else if(sce == 3)
  {
    P <- 5
    beta0 <- rep(c(-1,1), length = P)
    X <- matrix(NA, ncol = P, nrow = N)
    X[,1] <- rnorm(N)
    X[,2] <- rnorm(N)
    X[,3] <- rbinom(N, 1, 0.5)
    X[,4] <- rexp(N, 1)
    X[,5] <- rexp(N, 1)
  }else if(sce == 4)
  {
    P <- 5
    beta0 <- rep(c(-1,1), length = P)
    X <- matrix(NA, ncol = P, nrow = N)
    X[,1] <- rnorm(N)
    X[,2] <- rnorm(N)
    X[,3] <- rnorm(N)
    X[,4] <- rnorm(N)
    X[,5] <- rnorm(N)
  }else{
    P <- 10
    beta0 <- rep(c(-1,1), length = P)
    X <- matrix(NA, ncol = P, nrow = N)
    X[,1] <- rnorm(N)
    X[,2] <- rnorm(N)
    X[,3] <- rbinom(N, 1, 0.5)
    X[,4] <- rexp(N, 1)
    X[,5] <- rexp(N, 1)
    X[,6] <- rnorm(N)
    X[,7] <- rnorm(N)
    X[,8] <- rbinom(N, 1, 0.5)
    X[,9] <- rexp(N, 1)
    X[,10] <- rexp(N, 1)
  }

  z0 <- X%*%beta0
  r0=exp(abs(z0)^(a)*sign(z0))
  T <- rweibull(N,shape=2,scale=sqrt(r0))

  if(cc == 0)
  {
    C <- rep(Inf, N)
  }else{
    C <- rexp(N, cc)
  }

  Y <- pmin(T,C)
  E <- T<=C
  E[Y>tau] <- 0
  Y <- (pmin(Y,tau))

  ldata <- data.frame(X = X, Y = Y, E = E)
  if(res == 1)
  {
    ldata
  }else{
    mean(E) - cr0
  }
}

calPara <- function(cr0 = 0.75, a = 1/3, sce = 1)
{
  ldata <- genData(cc = 0, tau = Inf, sce = sce, cr0 = 0, N = 1e7, res = 1)
  tau <- floor(quantile(ldata$Y, probs = 0.95))
  cc <- uniroot(genData, lower = 0, upper = 10,
                sce = sce, N = 1e6, res = 2, cr0 = cr0, tau = tau)$root
  P <- dim(ldata)[2]-2
  beta0 <- rep(c(-1,1), length = P)
  Z0 <- round((as.matrix(ldata)[,1:P])%*%beta0,10)
  zz <- quantile(Z0, probs = seq(from = 0.01, to = 0.8, by = 0.01))
  tt <- seq(from = 0, to = tau, length = 100)

  ltau <- tau^2
  ltt0 = tt^2/ltau
  #
  rzz0 = exp(abs(zz)^(a)*sign(zz))/ltau

  list(cc = cc,
       tau = tau,
       zz = zz,
       tt = tt,
       ltt0 = ltt0,
       rzz0 = rzz0,
       P = P)
}

do.sim <- function(N, cr0, sce, a, cc, tau, zz, tt, rzz0, ltt0)
{
  ldata <- genData(cc = cc, tau = tau, sce = sce, res = 1, N = N, a = a)
  fit <- SIM(ldata, 10, zz, tt)
  mseL <- mean((fit$Ltt - ltt0)^2)
  mser <- mean((fit$rzz - rzz0)^2)
  c(fit$beta, -fit$betacox, mseL, mser)
}

library(Iso)
library(survival)
library(fastmatch)
library(SQUAREM)
library(doBy)
library(doParallel)
library(doRNG)
cl <- makeCluster(8)
registerDoParallel(cl)

set.seed(2020)

# 1000 iterations were considered in the simulations
B <- 1000
# a is defined in section 3 of the manuscript; it can be 1, 1/3, 1/5
a <- 1
# censoring rate is 1-cr0
cr0 <- 0.75
# scenario 2 in the manuscript
sce <- 2
# sample size is 800
N <- 800

para <- calPara(cr0=cr0, a=a, sce=sce)
cc <- para$cc
tau <- para$tau
zz <- para$zz
tt <- para$tt
ltt0 <- para$ltt0
rzz0 <- para$rzz0
P <- para$P

mat <- matrix(ncol = P*2+2, nrow = B)
err <- foreach(j = 1:B, .combine = rbind,
               .packages = c("survival","Iso","doBy",
                             "fastmatch","SQUAREM")) %dorng%
  {

    mat[j,] <- do.sim(N = N, cr0 = cr0, sce = sce, a = a, cc = cc, tau = tau,
                               zz = zz, tt = tt, rzz0 = rzz0, ltt0 = ltt0)
   }

save(err, file = paste0("res",N,"_",sce,"_",cr0,"_",a,".RData"))

stopCluster(cl)
